C
C ******************************************************************
C  
C  SUBROUTINE SEICTRI(P,ZA,ZB,ZAA,ZBB,ZBNX1,ZBNXN,ZBBNX1,ZBBNXN,ZE,NX)
C                                                          
C    IMPLICIT COMPLEX TRIDIAGONAL SOLVER:        
C                                                          
C            P  -   NX  (WAVEFIELD VECTOR / INPUT & OUTPUT)
C           ZA  -   NX                  
C           ZB  -   NX                                      
C          ZAA  -   NX                                 
C          ZBB  -   NX                                  
C        ZBNX1  -   SCALAR                                        
C        ZBNXN  -   SCALAR                                         
C       ZBBNX1  -   SCALAR                                   
C       ZBBNXN  -   SCALAR                                   
C           ZE  -   NX  (WORK ARRAY)
C
C
C    SOLVES THE FOLLOWING FIRST-ORDER LINEAR DIFFERENCE EQUATION:
C
C                                                           (K+1)
C    ___                                             ___    __  __ 
C   |                                                   |  |      |
C   | ZBNX1   ZA(2)     0      ...                      |  | P(1) |
C   |                                                   |  |      |
C   | ZA(1)   ZB(2)   ZA(3)    0                        |  | P(2) |
C   |                                                   |  |  .   |
C   |   0     ZA(2)   ZB(3)   ZA(4)                     |  |  .   |
C   |                                                   |  |  .   |
C   |                          ...                      |  |      |
C   |                                                   |  |      |
C   |                                 ZA(NX-1)     0    |  |  .   |
C   |                                                   |  |  .   |
C   |                                 ZB(NX-1)   ZA(NX) |  |P(NX-1)
C   |                                                   |  |      |
C   |                            0    ZA(NX-1)   ZBNXN  |  | P(NX)|
C   |___                                             ___|  |__  __|
C
C
C                                                              (K)
C      _______                                        ____    __  __ 
C     |                                                   |  |      |
C     |ZBBNX1   ZAA(2)    0       ...                     |  | P(1) |
C     |                                                   |  |      |
C     |ZAA(1)   ZBB(2)   ZAA(3)    0                      |  | P(2) |
C     |                                                   |  |  .   |
C     |   0     ZAA(2)   ZBB(3)  ZAA(4)                   |  |  .   |
C     |                                                   |  |  .   |
C     |                          ...                      |  |      |
C  =  |                                                   |  |      |
C     |                                ZAA(NX-1)     0    |  |  .   |
C     |                                                   |  |  .   |
C     |                                ZBB(NX-1)  ZAA(NX) |  |P(NX-1)
C     |                                                   |  |      |
C     |                            0   ZAA(NX-1)  ZBBNXN  |  | P(NX)|
C     |___                                           _____|  |__  __|
C
C
C    NOTE THAT WITH THIS INDEXING SCHEME THE INDICES LINE UP VERTICALLY
C    IN THE MATRIX COLUMNS. THIS IS BECAUSE THE PHYSICAL CONSTANTS
C    WHICH DEFINE THE MATRIX ENTRIES DEPEND ON SPATIAL LOCATION.
C
C     
      SUBROUTINE SEICTRI( P, ZA, ZB, ZAA, ZBB,
     1			    ZBNX1, ZBNXN, ZBBNX1, ZBBNXN,
     2           	    ZE, NX )
C  
      implicit none
      integer  nx, i, jnx1, ixm1, ixp1
      complex      p(  nx)                                  
      complex     za(0:nx+1)                                
      complex     zb(  nx)                                  
      complex    zaa(  nx)                                  
      complex    zbb(  nx)                                  
      complex     ze(  nx)                                  
      complex  den, zf, zbnx1, zbnxn, zbbnx1, zbbnxn              

C  Construct right-hand side of tri-d system
C  First element
	ze(1) = zbbnx1*p(1) + zaa(2)*p(2)
C  Mid-range elements
	do i=2,nx-1
		ze(i) = zaa(i-1)*p(i-1) + zbb(i)*p(i) + zaa(i+1)*p(i+1)
	enddo
C  Last element
	ze(nx) = zaa(nx-1)*p(nx-1) + zbbnxn*p(nx)

C
C  Call LIBSCI complex tri-d solver
	call cdtsol( nx, za(0), zb, za(2), 1, ze, 1 )

C
C  Move solution back
	do i=1,nx
		p(i) = ze(i)
	enddo

	return
	end
