C*----------------------------------------------------------------------
C*
C*  NAME:
C*    dmocp, dmoca, dmocb, dmocc - Seismic Processing Benchmark 
C*        process DMOC
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call dmocp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call dmoca( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call dmocb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call dmocc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    Reads data from seismic benchmark format files
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1        10    Debug flag, set to 1 for debug output
C*      2     11-20    Velocity for dip moveout application
C*      3     21-30    CMP Separation
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./dmoc.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     dmocp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'DMOC' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine dmocp( ldim, maxtrc, nra, nsa, abort, ipr )
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
	include 'coord.inc'
	include 'group.inc'
	include 'dmoc.inc'

C
C  Local declarations:
C
C  parameter record buffer
	character*80 buf
C  parameter record index
	integer index

C
C  Start of executable statements
C

C  Get parameter record
	index = 1
	call getparm('DMOC',index,buf,ier)
	if (ier.ne.0) then
		write (ipr,*) 'No parameter records for process DMOC'
		abort = .true.
		return
	endif

C  Read velocity and cmp separation
	read (buf,'(7x,i3,2f10.0)') idbug,vdmo,delx
	if (vdmo.eq.0) then
		write (ipr,*) 'DMOC: DMO Velocity must be non-zero'
		abort = .true.
		return
	endif
	if (delx.eq.0) delx = gdx

C  Check for proper data type
	if (gtype.ne.'xoff    ') then
		write (ipr,*) 'DMOC: data must be common offset (xoff)'
		abort = .true.
		return
	endif

	write (ipr,9010) vdmo,delx
 9010 format (/,' Process DMOC    : Dip Moveout Correction',/,
     +        ' DMO Velocity    : ',f8.1,/,
     +        ' CMP Separation  : ',f8.2)

C  Convert velocity to ft/sample
	vdmo = vdmo * 0.001*msr

C  Set memory requirements
	ldim = ns
	maxtrc = ntpg
	nra = 0
C  Need work space to hold aperture and common offset section
	nsa = ns*(ntpg+1)

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     dmoca, dmocb, dmocc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for dmocing Seismic Benchmark 
C*     Format files
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  The 'a' does nothing
C*----------------------------------------------------------------------

	subroutine dmoca( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine applies dmo
C*----------------------------------------------------------------------

	subroutine dmocb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	implicit none
C  Arguments
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort
C  Common blocks
	include 'coord.inc'
	include 'group.inc'
	include 'dmoc.inc'
C  Local variables
	integer i,j,k,ka
	real x,h

	ntro = ntri

C  Half-offset of this record
	h = 0.5*abs(sxyz(1,1) - rxyz(1,1))

C  Compute dmo aperture function
	ka = ns*ntri + 1
	call dmocapr( ns, vdmo, h, sa(ka) )
	if (idbug.ne.0) write (ipr,9010) h, (sa(ka+i),i=0,ns-1,100)
 9010 format (' Offset ',f8.1,' DMO Aperture: ',/,(10f7.1))

C  Return if offset is zero
	if (h.eq.0.) return

C  zero work array
	k = 0
	do j=1,ntri
		do i=1,ns
			k = k+1
			sa(k) = 0.0
		enddo
	enddo

C  Loop and integrate each trace
	do j = 1, ntri
C  Set trace location
		x = delx*(j-1)
C  Integrate over DMO ellipse
		call dmocint( delx, h, x, ntri, ns, sa, otr(1,j), 
     +			sa(ka) )
	enddo

C  Copy DMO section back to input
	k = 0
	do i=1,ns
		do j=1,ntri
			k = k+1
			otr(i,j) = sa(k)
		enddo
	enddo

	return
	end

C*----------------------------------------------------------------------
C*  'c' routine summarizes processing
C*----------------------------------------------------------------------
c
	subroutine dmocc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	call sysmsg('Process DMOC completed')
	ntro = 0

	return
	end

C*----------------------------------------------------------------------

	subroutine dmocint( delx, h, x, ntri, ns, sa, otr, xr )
C
C  Integrate input trace over DMO Ellipse into output array
C  real delx - cmp separation
C  real h,x - half-offset and midpoint location of input trace
C  integer ntri - number of input traces
C  integer ns - number of input samples
C  real sa(ntri,ns) - DMO accumulator array
C  otr(ns) - input seismic trace
C  xr(ns) - dmo aperture for each time sample
C
	implicit none
C  Arguments
	integer ns,ntri
	real delx,h,x,sa(ntri,ns),otr(ns),xr(ns)
C  Local variables
	integer i,j,k,ns1,ix1,ixn
	real h2,xm2,xs,dk

C  Inverse half-offset squared
	h2 = 1./(h*h)
C  Number of samples - 1
	ns1 = ns - 1

C  Loop over number of output samples
	do j=2,ns
C  Set integration range 
		ix1 = 1 + nint((x-xr(j))/delx)
		if (ix1.lt.1) ix1 = 1
		ixn = 1 + nint((x+xr(j))/delx)
		if (ixn.gt.ntri) ixn = ntri
C  Loop and sum
		do i=ix1,ixn,1
C  midpoint separation squared
			xm2 = (x - delx*(i-1))**2
C  Compute floating point sample to be summed
			xs = float(j-1)/sqrt(1. - xm2*h2)
C  Input sample number
			k = ifix(xs)
C  Set output value 
			if (k.lt.ns1) then
				dk = xs - k
				sa(i,j) = sa(i,j) + dk * otr(k+2) + 
     +					(1.-dk) * otr(k+1)
			endif
		enddo
	enddo

	return
	end
C
C

C*----------------------------------------------------------------------

	subroutine dmocapr( ns, vdmo, h, xr )
C
C  Set dmo aperture for each time sample
C  ns - number of samples
C  vdmo - dmo velocity in ft/sample (i.e. v * dt)	
C  h - half offset
C  xr(ns) - output dmo aperture half-width
C
	implicit none
C  Arguments
	integer ns
	real vdmo,h,xr(ns)	
C  Local variables
	integer j

	if (h.eq.0) then
		call vfill( 0.0, xr, 1, ns )
		return
	endif

	do j=1,ns
		xr(j) = 0.5*h / sqrt( 1.0 + (vdmo*(j-1)/(2.*h))**2 )
	enddo
	
	return
	end

C*----------------------------------------------------------------------
