C*----------------------------------------------------------------------
C*
C*  NAME:
C*    m3fkp, m3fka, m3fkb, m3fkc - Seismic Processing Benchmark 
C*        process M3FK
C*     
C*  SYNOPSIS:
C*    integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
C*    real otr(ldim,maxtrc), ra(nra), sa(nra)
C*    logical abort
C*    call m3fkp( ldim, maxtrc, nra, nsa, abort, ipr )
C*    call m3fka( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                abort, ipr )
C*    call m3fkb( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntri, ntro, abort, ipr )
C*    call m3fkc( ldim, maxtrc, otr, nra, ra, nsa, sa,
C*                ntro, abort, ipr )
C*
C*  DESCRIPTION: 
C*    3D constant velocity FK migration
C*
C*  PARAMETERS: 
C*    Record 1
C*    Field  Columns   Description
C*      1      7       debug flag
C*      2     11-20    migration velocity
C*      3     21-25    max frequency in Hz
C*      4     26-30    pass number (1=x, 2=y)
C*      5     31-35    padding factor in time direction
C*      6     36-40    padding factor in distance direction
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*     Include files:
C*       ./coord.inc
C*       ./group.inc
C*       ./para.inc
C*       ./m3fk.inc
C*
C*     Libraries:
C*       libbench.a
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        12/31/91 - C. Mosher, ARCO E&PT 
C*
C*      References:
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*   NAME:
C*     m3fkp
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   This process reads 'M3FK' parameters, checks syntax, and sets
C*   memory requirements for the processing phase
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*      Subroutines:
C*        getparm - read parameter record
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/06/90 - C. Mosher
C*
C*      References:
C*        SPARC System Documentation
C*
C*      Modified:
C*
C*----------------------------------------------------------------------
C
	subroutine m3fkp( ldim, maxtrc, nra, nsa, abort, ipr )
C
C  Arguments:
C
C  Output leading dimension of trace array required for the PROC step
	integer LDIM
C  Output max number of traces per group to be output during PROC step
	integer MAXTRC
C  Number of words of reserved storage required in PROC step
	integer NRA
C  Number of words of working storage required in PROC step
	integer NSA
C  Abort flag, set to .TRUE. by this process if errors detected
	logical ABORT
C  Print unit number that this routine should use
	integer IPR

C
C  Common blocks:
C
C  parameters for process M3FK
	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'm3fk.inc'

C
C  Local declarations:
C
C  parameter record buffer
	character*80 buf
C  parameter record index
	integer index

C
C  Start of executable statements
C

C  Get parameter record with velocity
	index = 1
	call getparm('M3FK',index,buf,ier)
	if (ier.ne.0) then
		write (ipr,*) 'No parameter records for process M3FK'
		abort = .true.
		return
	endif

C  Read velocity, max frequency, and padding factors
	read (buf,'(6x,i1,3x,5f5.0,4i5,f5.0)') idbug, tmax, fmax, vmig,
     +    dx, dy, ipadt, ipadx, ipady, lr, bw

C  Set increments in each direction
	if (tmax.eq.0.) tmax = msr*(ns-1)
	if (dx.eq.0.) dx = gdx
	if (dy.eq.0.) dy = gdy

C  Set transform lengths
	nt = 1 + nint(tmax/float(msr))
	call seiftm(nt,magn,n2)
	magn = magn + ipadt
	n2 = 2**magn
	nyq = n2/2 + 1
	call seiftm(ntpg,magm,m2)
	magm = magm + ipadx
	m2 = 2**magm
	myq = m2/2 + 1
	call seiftm(ngpl,magl,l2)
	magl = magl + ipady
	l2 = 2**magl
	lyq = l2/2 + 1

C  Number of frequencies to process
	dt = 0.001*float(msr)
	df = 1./(dt*n2)
	if (fmax.ne.0.) then
		nf = 1 + nint(fmax/df)
	else
		nf = n2/2
	endif

C  Force number of frequencies to multiple of number of processors
	nfp = nf/nodes
	if (nfp*nodes .lt. nf) then
		nfp = nfp + 1
		nf = min( n2/2, nfp*nodes )
	endif
	fmax = df*(nf-1)

C  Depth/Distance ratios
	zmax = 0.5*vmig*(0.001*tmax)
	xmax = dx*m2
	ymax = dy*l2
	rzx = zmax/xmax
	rzy = zmax/ymax

C  Interpolation operator length and bandwidth
	lr = 2*(lr/2)
	if (lr.le.0 .or. lr.gt.12) lr = 6
	if (bw.le.0. .or. bw.gt.0.9) bw = 0.8
				
	write (ipr,9010) tmax,fmax,vmig,zmax,xmax,ymax,n2,m2,l2,lr,bw
 9010 format (' Process M3FK    : 3D FK Migration',/,
     +        ' Max time        : ',f8.1,/,
     +        ' Max frequency   : ',f8.1,/,
     +        ' Migration vel   : ',f8.1,/,
     +        ' Max z,x,y       : ',3f8.1,/,
     +        ' Transform size  : ',3i8,/,
     +        ' Intrp opr len   : ',i8,/,
     +        ' Intrp bandwidth : ',f8.2)

C  Quit if number of nodes is not power of 2
	call seiftm( nodes, mag, nn )
	if (nn.ne.nodes) then
		write (*,*) 'M3FK Error: number of nodes must be power of 2'
		abort = .true.
		return
	endif

C  Set memory requirements
	m2p = m2/nodes
	l2p = l2/nodes
C  2D frame size
	lframe = 2*nf*m2
C  Transform volume
	nra = lframe*l2p

C  Work space for 1 trace
	nsa = nyq
C  Need work space to hold one 'tile' for distributed transposes
	if (nodes.gt.1) then
		nsa = max( nsa, m2*l2p*nfp, l2*m2p*nfp, nf*m2p*l2p ) 
	endif

C  Work space is complex
	nsa = 2*nsa

C  Check to be sure there is enough work space for interp routines
	nsa = max(nsa, 4*nf + (lr+2)*101 + 2*lr)

	return
	end

C*----------------------------------------------------------------------
C*   NAME:
C*     m3fka, m3fkb, m3fkc
C*     
C*----------------------------------------------------------------------
C*   DESCRIPTION:
C*
C*   Processing phase routines for Seismic Benchmark process M3FK
C*
C*----------------------------------------------------------------------
C*   DEPENDENCIES:
C*
C*----------------------------------------------------------------------
C*   HISTORY:
C*
C*      Written:
C*        01/01/92 - C. Mosher
C*
C*      References:
C*        Seismic Benchmark Users Guide
C*
C*      Modified:
C*
C*----------------------------------------------------------------------

C*----------------------------------------------------------------------
C*  Work phase initialization
C*----------------------------------------------------------------------

	subroutine m3fka( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		abort, ipr )
	integer ldim, maxtrc, nra, nsa, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'group.inc'
	include 'para.inc'
	include 'm3fk.inc'

	if (idbug.ne.0 .and. node.eq.master) then
		idbug = ipr
	else
		idbug = 0
	endif

C  Initialize input/output 2D frame counters
	jframe = 0
	kframe = 0

C  Initialize data volume
	call vfill( 0.0, ra, 1, lframe*l2p )

C  Initialize compute and communication time
	compt = 0.0
	commt = 0.0

	return
	end

C*----------------------------------------------------------------------
C*  The 'b' routine processes data
C*----------------------------------------------------------------------

	subroutine m3fkb( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntri, ntro, abort, ipr )
	integer ldim, maxtrc, nra, nsa, ntri, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'coord.inc'
	include 'group.inc'
	include 'm3fk.inc'

C  pointer into array of stored data
	k = jframe * lframe + 1
	jframe = jframe + 1

C  Forward 1D fft to (f,x)
	t0 = etimer(0)
	call seimrcf( ldim, ntri, otr, nf, ra(k), n2, ns, nf, sa )
	compt = compt + etimer(1)

C  Transpose to (x,f)
	t0 = etimer(0)
	call tran21c( nf, m2, ra(k) )
	commt = commt + etimer(1)

C  Complex transform over x direction to (kx,f)
	idir = 1
	t0 = etimer(0)
	call m3fk_mcft( m2, nf, ra(k), idir ) 
	compt = compt + etimer(1)

C  No traces passed out
	ntro = 0

	return
	end

C*----------------------------------------------------------------------
C*  Migrate and output migrated traces
C*----------------------------------------------------------------------
c
	subroutine m3fkc( ldim, maxtrc, otr, nra, ra, nsa, sa, 
     +		ntro, abort, ipr )
	implicit none
	integer ldim, maxtrc, nra, nsa, ntro, ipr
	real otr(ldim,maxtrc), ra(nra), sa(nsa)
	logical abort

	include 'coord.inc'
	include 'group.inc'
	include 'para.inc'
	include 'm3fk.inc'

	integer k, idir
	real t0, etimer, bytesm, flopsm, flops1, flops2, flops3

C  Set group number for output
	igroup = node*l2p + kframe + 1

C  Migrate data if first time through
	if (kframe.eq.0) then
		call m3fk_fkmig( ra, sa )
C  If up to number of lines, quit
	else if (igroup.gt.ny) then
		ntro = 0
		if (node.eq.master) then
C  Master computes estimated flop count
			flops1 = float(l2)*float(m2)*float(n2)*alog(float(n2))
			flops2 = float(nf)*float(l2)*float(m2)*alog(float(m2))
			flops3 = float(nf)*float(m2)*float(l2)*alog(float(l2))
C  Fudge factor (empirically determined from Cray HPM)
C  Note that forward/inverse real transform is 1/2 complex 
			flopsm = 1.e-6 * 6.5 * (flops1 + 2.*flops2 + 2.*flops3)
C  Mbytes/sec moved in global forward/inverse transpose operation
			bytesm = 4.e-6*float(nf)*float(m2)*float(l2)/commt
			write (ipr,9010) compt,commt,compt/commt,
     +        flopsm/compt,flopsm/(compt+commt),bytesm
 9010   format (' Process M3FK',/,
     +			' 3D FFT Statistics',/,
     +			' Computation    time (sec)    : ',f10.2,/,
     +			' Communication  time (sec)    : ',f10.2,/,
     +			' Compute/Communicate Ratio    : ',f10.2,/,
     +			' Raw Mflops (no communication): ',f10.2,/,
     +			' Effective Mflops (comp+comm) : ',f10.2,/,
     +			' Global transpose rate (Mb/s) : ',f10.2)
		endif
		call sysmsg('Process M3FK completed')
		return
	endif

C  Set pointer into migrated data volume
	k = lframe*kframe + 1
	kframe = kframe + 1

C  Complex inverse transform over x direction to (x,f)
	idir = -1
	t0 = etimer(0)
	call m3fk_mcft( m2, nf, ra(k), idir ) 
	compt = compt + etimer(1)

C  Transpose to (f,x)
	t0 = etimer(0)
	call tran21c( m2, nf, ra(k) )
	commt = commt + etimer(1)

C  Inverse fft to extract data
	t0 = etimer(0)
	call seimcrf( ldim, ntpg, otr, nf, ra(k), n2, ns, nf, sa )
	compt = compt + etimer(1)
	ntro = ntpg
	iline = 1

	return
	end
C
C*----------------------------------------------------------------------
C*  Support subroutines
C*----------------------------------------------------------------------
C
C*----------------------------------------------------------------------
C
	subroutine m3fk_fkmig( ra, sa )
C
C  Complete 3D fft and interpolate from f to kz
C
C  Input data is stored as ( kx, f, y ) with y distributed
C  across the nodes.  Use local and distributed transpose
C  routines to bring y to the front, perform 3rd fft, and
C  then interpolate from f to kz.  Then, use transpose operations
C  to return the data to ( kx, kz, y ).
C
C  Size of the data is defined in 'm3fk.inc', by the values:
C  nf - number of frequencies
C  m2 - number of x values
C  l2 - number of y values
C  nfp - nf/nodes
C  m2p - m2/nodes
C  l2p - l2/nodes
C
C  Note that for distributed operations, there are nfp, m2p, or
C  l2p 2D arrays on each processor, depending on whether f, x, or y
C  is the last dimension of the distributed array.
C
	implicit none
	complex ra(*), sa(*)
	include 'para.inc'
	include 'm3fk.inc'
	integer idir
	real t0,etimer

	if (idbug.ne.0) write (idbug,*) 'M3FK: transpose'
	t0 = etimer(0)
C  Stored data volume is ( kx, f, y ), with y spread across nodes.
C  Use transpose operations to bring y to the front.
C  ( kx, f, y )
	call tran213c( m2, nf, l2p, ra )
C  ( f, kx, y )
	call dtran132c( nf, m2, l2, ra, sa )
C  ( f, y, kx )
	call tran213c( nf, l2, m2p, ra )
C  ( y, f, kx )
	commt = commt + etimer(1)

	if (idbug.ne.0) write (idbug,*) 'M3FK: last fft'
C  Forward transform to ( ky, f, kx )
	idir = 1
	t0 = etimer(0)
	call m3fk_mcft( l2, m2*nfp, ra, idir )
	compt = compt + etimer(1)

	if (idbug.ne.0) write (idbug,*) 'M3FK: interpolate'
C  Bring frequencies to the front for interpolation
C  ( ky, f, kx )
	t0 = etimer(0)
	call tran213c( l2, nf, m2p, ra )
	commt = commt + etimer(1)
C  ( f, ky, kx )

C  Interpolate from ( f, ky, kx ) to ( kz, ky, kx )
CCM Leave out interpolation time from total, count FFT flops only
	call m3fk_fkint( nf, n2, l2, m2, ra, rzy, rzx,
     +    lr, bw, sa )

C  Bring ky to the front and inverse transform
C  ( kz, ky, kx )
	t0 = etimer(0)
	call tran213c( nf, l2, m2p, ra )
	commt = commt + etimer(1)
C  ( ky, kz, kx )
	if (idbug.ne.0) write (idbug,*) 'M3FK: inverse fft'
	idir = -1
	t0 = etimer(0)
	call m3fk_mcft( l2, nf*m2p, ra, idir )
	compt = compt + etimer(1)
C  ( y, kz, kx )

	if (idbug.ne.0) write (idbug,*) 'M3FK: inverse transpose'
	t0 = etimer(0)
C  Bring kx to the front and return.
C  ( y, kz, kx )
	call tran213c( l2, nf, m2p, ra )
C  ( kz, y, kx )
	call dtran132c( nf, l2, m2, ra, sa )
C  ( kz, kx, y )
	call tran213c( nf, m2, l2p, ra )
C  ( kx, kz, y )
	commt = commt + etimer(1)
	if (idbug.ne.0) write (idbug,*) 'M3FK: transpose complete'

C  That's it.  Whew.

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine m3fk_mcft( m2, nf, ra, idir )
C
C  Multiple complex fft
C
C  m2 - transform length
C  nf - numbef of transforms
C  ra(m2,nf) - input/output transform array
C  idir - transform direction, 1=forward, -1=inverse
C
	complex ra(m2,nf)

	do j=1,nf
		call seicft( m2, ra(1,j), idir )
	enddo

	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine m3fk_fkint( nf, n2, m2, l2, ra, rzx, rzy,
     +    lr, bw, sa )
c
c  Interpolation for 3D fk migration
c
c  nf - number of frequencies to process
c  n2 - transform length in time direction
c  m2 - transform length in x direction
c  l2 - transform length in y direction
c  ra - input/output transform array
c  rzx - depth to x distance ratio 
c  rzy - depth to y distance ratio 
c  lr - interpolation operator length
c  bw - interpolation operator bandwidth
c  sa - work array, length 2*(nf+lr) + 101*lr + 2*nf + 202
c
	implicit none
c  Argument types
	integer nf, n2, m2, l2, lr
	complex ra(nf,m2,*)
	real rzx, rzy, bw, sa(*)
c  Local types
	integer ksa, kop, knrm, kmod, ksr, ksi
	integer myq, lyq, j, k, kk, kx, ky, ky1, kyn

c  Work array pointers
	ksa = 1
c  Sync operators
	kop = ksa
	ksa = ksa + 101*lr
c  Complex sync modulation factors
	kmod = ksa
	ksa = ksa + 2*nf
c  Complex sync normalization factors
	knrm = ksa
	ksa = ksa + 202

c  Genereate operators
	call syncini( lr, bw, sa(kop), sa(ksa) )
	call synccnm( bw, n2, nf, sa(kmod), sa(knrm) )

c  Work arrays for interpolation
	ksr = ksa
	ksa = ksa + nf + lr
	ksi = ksa

c  Nyquist samples
	myq = m2/2 + 1
	lyq = l2/2 + 1
  
C Get range of wavenumbers on this processor
	call jrange( l2, ky1, kyn )
c  Index into local array
	k = 1
c  Loop over global wavenumber index in y direction
	do kk = ky1,kyn
c  if greater than nyquist, convert to positive wavenumber index
		if (kk.gt.lyq) then
			ky = l2 - kk + 2
		else
			ky = kk
		endif
c  loop over x wavenumbers
		do j=1,m2
			kx = j
			if (kx.gt.myq) kx = m2 - kx + 2
c  interpolate
			call m3fk_kzint( nf, ra(1,j,k), rzx, rzy, kx, ky, 
     +        lr, sa(kop), sa(kmod), sa(knrm), sa(ksr), sa(ksi) )
		enddo
c  Increment pointer into local array
		k = k + 1
	enddo
	return
	end
C
C*----------------------------------------------------------------------
C
	subroutine m3fk_kzint( nf, ra, rzx, rzy, kx, ky, 
     +    lr, cof, cmod, cnrm, sr, si )
C
C  post-stack 3D fk migration interpolation
C  nf - number of values to interpolate
C  ra - input vector of omega values, output interpolated kz values
C  rzx - midpoint aspect ratio (zmax/xmax)
C  rzy - line aspect ratio (zmax/ymax)
C  kx - midpoint wavenumber index
C  ky - line wavenumber index
C  lr,cof - interpolation operators
C  cmod,cnrm - modulation/normalization factors
C  sr,si - work arrays, length nf+lr
C
	implicit none
C  Argument types
	integer nf, kx, ky, lr
	real rzx, rzy, sr(*), si(*), cof(lr,101)
	complex ra(nf), cmod(nf), cnrm(101), ctemp
C  Local types
	integer i, k, l, m, lr2, iw
	real w, vr, vi, rzkx2, rzky2
C
C  loop and move real and imag parts of input array to work arrays
C  apply demodulation factors to reduce phase variations
C  pad ends with zeros for interpolation wrap-around
	lr2 = lr/2
	k = 1
	do i=1,lr2-1
	    sr(k) = 0.
	    si(k) = 0.
	    k = k+1
	enddo
	do i=1,nf
	    ctemp = ra(i) * conjg(cmod(i))
	    sr(k) = real(ctemp)
	    si(k) = aimag(ctemp)
	    k = k+1
	enddo
	do i=1,lr2
	    sr(k) = 0.
	    si(k) = 0.
	    k = k+1
	enddo

C  set aspect ratio factors
	rzkx2 = (rzx*float(kx-1))**2
	rzky2 = (rzy*float(ky-1))**2
C  loop and interpolate
	do i = 2,nf
C  compute floating point frequency sample number
	    w = sqrt( float(i-1)**2 + rzkx2 + rzky2 )
C  set index of lower bracketing value
	    iw = ifix(w)
	    if (iw.lt.nf) then
C  set interpolation operator number
	        l = 1 + nint(100.*(w-iw))
	        vr = 0.
	        vi = 0.
	        k = iw
C  loop and convolve
	        do m=1,lr
	            k = k+1
	            vr = vr + sr(k)*cof(m,l)
	            vi = vi + si(k)*cof(m,l)
			enddo
C  set value - remodulate using interpolated frequency index
	        ra(i) = cmod(iw+1) * cnrm(l) * cmplx(vr,vi)
	    else
	        ra(i) = (0.,0.)
	    endif
	enddo

	return
	end
