/*--------------------------------------------------------------------*/
/*  NAME:
/*    fetch.c - Parameter keyword file I/O
/*
/*  VERSION:
/*    Initial version, simple keyword=value I/O support
/*
/*  Routines defined in this module:
/*    ifetch - return integer from matching keyword
/*    rfetch - return float (real) from matching keyword
/*    fetch - return character string from matching keyword
/*  Additional Fortran interfaces:
/*    opnprm - open paramter file
/*    clsprm - close paramter file
/*     
/*--------------------------------------------------------------------*/
/*
/*  HISTORY:
/*
/*     Written:
/*       10/08/93 - C. Mosher, ARCO E&PT 
/*
/*     References:
/*
/*     Modified:
/*
/*--------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>

/*--------------------------------------------------------------------*/
/*  NAME:
/*    ifetch, isfetch - return integer from matching keyword
/*     
/*  SYNOPSIS:
/*    FILE *fp;
/*    char *buf;
/*    char *keyword;
/*    int *ival;
/*    int ier;
/*    ier = ifetch( fp, keyword, ival );
/*    ier = isfetch( buf, keyword, ival );
/*
/*    Retrieves a 'keyword=integer' record from a parameter file
/*    or character buffer.
/*    Return value is zero on success, negative on errors:
/*    -1 = no match found
/*    -2 = parsing error
/*
/*--------------------------------------------------------------------*/

int ifetch(fp,keyword,ival)
FILE *fp;		/* file pointer for keyword file */
char *keyword;	/* input keyword to be matched */
int *ival;		/* returned integer value if match found */
{
	int ier;
	char buf[256];

	ier = fetch(fp,keyword,buf);
	if (ier > 0) {
		sscanf(buf,"%d",ival);
		ier = 0;
	}
	return(ier);

}

/*--------------------------------------------------------------------*/

int isfetch(buf,keyword,ival)
char *buf;		/* character buffer */
char *keyword;	/* input keyword to be matched */
int *ival;		/* returned integer value if match found */
{
	int ier;
	char string[256];

	if (sfetch(buf,keyword,string) > 0 ) {
		if (sscanf(string,"%d",ival) == 1)
			return(0);
		else
			return(-1);
	}
	return(-1);

}
		

/*--------------------------------------------------------------------*/
/*  NAME:
/*    rfetch,rsfetch - return float from matching keyword
/*     
/*  SYNOPSIS:
/*    FILE *fp;
/*    char *buf;
/*    char *keyword;
/*    float *val;
/*    int ier;
/*    ier = rfetch( fp, keyword, val );
/*    ier = rsfetch( buf, keyword, val );
/*
/*    Retrieves a 'keyword=float' record from a parameter file
/*    or character buffer.
/*    Return value is zero on success, negative on errors:
/*    -1 = no match found
/*    -2 = parsing error
/*
/*--------------------------------------------------------------------*/

int rfetch(fp,keyword,val)
FILE *fp;		/* file pointer for keyword file */
char *keyword;	/* input keyword to be matched */
float *val;		/* returned float value if match found */
{
	int ier;
	char buf[256];
	double dval;

	ier = fetch(fp,keyword,buf);
	if (ier > 0) {
		sscanf(buf,"%f",val);
		ier = 0;
	}
	return(ier);

}

/*--------------------------------------------------------------------*/

int rsfetch(buf,keyword,val)
char *buf;		/* character buffer */
char *keyword;	/* input keyword to be matched */
float *val;		/* returned float value if match found */
{
	char string[256];

	if (sfetch(buf,keyword,string) > 0 ) {
		if (sscanf(string,"%f",val) == 1)
			return(0);
		else
			return(-1);
	}
	return(-1);

}
		

/*--------------------------------------------------------------------*/
/*  NAME:
/*    fetch, sfetch - return string from matching keyword
/*     
/*  SYNOPSIS:
/*    FILE *fp;
/*    char *buf;
/*    char *keyword;
/*    char *string;
/*    int ier;
/*    ier = fetch( fp, keyword, string );
/*    ier = sfetch( buf, keyword, string );
/*
/*    Retrieves a 'keyword=string' record from a parameter file
/*    or character buffer.
/*    Return value is string length on success, negative on errors:
/*    -1 = no match found
/*    -2 = keyword found, but no equal sign
/*
/*--------------------------------------------------------------------*/

int fetch(fp,keyword,string)
FILE *fp;		/* file pointer for parameter file */
char *keyword;	/* input keyword to be searched for */
char *string;	/* output string matching keyword */
{
#	define MAXLEN 256
	char buf[MAXLEN];
	char *fgets();
	int len;

/* Start from beginning of parameter file */
	rewind(fp);

/* Loop and read lines from parameter file, searching for  */
/* a match to the keyword string */
	while ( fgets( buf, MAXLEN, fp ) != (char *)NULL ) {
/* check for comment line */
		if (*buf == '#') continue;
/* Check for keyword match on this line */
		len = sfetch( buf, keyword, string );
		if (len) return(len);
/* no keyword match on this line, get next line from file */
	}
   
/* No match in file, return with end of file condition */
	return(-1);
}

/*--------------------------------------------------------------------*/

int sfetch( buf, keyword, string )
/* 
	Return substring matching keyword
*/
char *keyword, *buf, *string;
{
	int klen;
	char *strnstr();

/* Return on zero length keyword */
	if ( (klen=strlen(keyword)) == 0 ) return(0);

/* Look for substring match to keyword in input string */
	while (buf = strnstr(buf,keyword,klen)) {
/* found keyword match, look for equal sign after white space */
		buf += klen;
		buf += strspn(buf," \t");
/* Continue search if equal sign not found after keyword */
		if (*buf != '=') continue;
/* equal sign found, move past white space (blank,tab) */
		buf++;
		buf += strspn(buf," \t");
/* Check to see if string was quoted */
		if ( *buf == '"' )
/* if quoted, move chars up to next quote or end of string */
			gstrccpy(string,++buf,'"');
		else
/* not quoted, move chars until blank, comma, tab, newline, or null */
			gstrwcpy(string,buf);
/* return with string length */
		return(strlen(string));	
	}

/* No match */
	return(0);
}

/*--------------------------------------------------------------------*/
/* Internal support routines
/*--------------------------------------------------------------------*/

gstrccpy(to,from,c)
/*
	Copy characters until delimeter char or end of string encountered 
*/
char *to,*from,c;
{
	while ( (*from != c) && (*from != '\0') ) 
		*to++ = *from++;

	*to = '\0';
	return;
}

/*--------------------------------------------------------------------*/

gstrwcpy(to,from)
/*
	Copy characters until white space or a comma is encountered
*/
char *to,*from;
{
	while ( (*from != '\0') &&
			(*from !=  ' ') &&
			(*from !=  ',') &&
			(*from != '\t') &&
			(*from != '\n') )
		*to++ = *from++;

	*to = '\0';
	return;
}
 
/*--------------------------------------------------------------------*/

strnwcp(to,from,n)
/*
	Copy up to n characters until white space is encountered.
	Output string will have up to n+1 chars including null.
*/
char *to,*from;
int n;
{
	while ( (n--) &&
			(*from != '\0') &&
			(*from !=  ' ') &&
			(*from !=  ',') &&
			(*from != '\t') &&
			(*from != '\n') )
		*to++ = *from++;

	*to = '\0';
	return;
}
 
/*--------------------------------------------------------------------*/
 
int lenstrw( s, n )
/*
	Return length of string up to first white space encountered.
*/
char *s;		/* input string */
int n;			/* max length to check */
{
	int i;

	i = 0;
	while ( (n--) &&
			(*s != '\0') &&
			(*s !=  ' ') &&
			(*s !=  ',') &&
			(*s != '\t') &&
			(*s != '\n') ) {
		*s++;
		i++;
	}

	return(i);
}
 
/*--------------------------------------------------------------------*/

char *strnstr(buf,string,n)
/*
	Return pointer to first occurence of 'n' characters from 'string'
	in 'buf'.  Returns null if no match.
*/
char *buf, *string;
int n;
{
	while ( *buf ) {
		if (strncmp(buf,string,n) == 0) return(buf);
		buf++;
	}

	return((char *)0);
}
